import json
import os

from houston.exceptions import HoustonClientError, HoustonException
from houston.interface import InterfaceRequest

try:
    basestring
except NameError:
    basestring = str


class Houston:
    def __init__(self, plan, api_key=None):
        """
        :param plan: Can be either:
                        - string: name of an existing plan to be loaded
                        - dict: plan definition, see Plan docs for example
                        If a plan definition is provided, ensure it is loaded with load_plan to make it
                        available to other instances
        :param string api_key: Api key provided from https://callhouston.io, see account information. Can also be set as
                        environment variable HOUSTON_KEY
        """
        if api_key is None:
            api_key = os.getenv("HOUSTON_KEY", None)
        if api_key is None:
            raise ValueError(
                "No API key was found. Provide 'api_key' parameter or set the "
                "HOUSTON_KEY environment variable."
            )
        self.key = api_key
        self.interface_request = InterfaceRequest(key=api_key)
        self.base_url = "https://callhouston.io/api/v1"

        if isinstance(plan, str):
            self.plan = self.get_plan(plan)
        else:
            self.plan = plan

        if "name" not in self.plan:
            raise HoustonClientError(
                "Sorry, this plan is not formatted correctly - must contain a name"
            )

    def save_plan(self):
        """Sends selected plan to Houston server"""
        self.interface_request.request(
            method="post", uri=self.base_url + "/plans", data=json.dumps(self.plan)
        )

    def delete_plan(self, safe=False):
        """Removes selected plan from Houston server

        :param boolean safe: Ignore exception raised by invalid request i.e. plan doesn't exist, True = ignore
        """
        plan_name = self.plan["name"]
        self.interface_request.request(
            "delete", uri=self.base_url + "/plans/" + plan_name, safe=safe
        )

    def get_plan(self, plan_name=None):
        """Get saved plan detail from Houston server

        :param string plan_name: [Optional] name of plan to retrieve details of
        :return dict: plan detail
        """
        if plan_name is None:
            plan_name = self.plan["name"]

        status_code, response = self.interface_request.request(
            "get", uri=self.base_url + "/plans/" + plan_name
        )
        return response

    def create_mission(self):
        """Creates a new instance of a mission
        :return string: new mission instance id or False if error
        """
        status_code, response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions",
            data=json.dumps({"plan": self.plan["name"]}),
        )
        if "id" not in response:
            raise HoustonException(
                "Create mission operation did not return a mission_id, please retry"
            )

        return response["id"]

    def start_stage(self, stage_name, mission_id, retry=3, ignore_dependencies=False):
        """Starts Houston stage, returns current stage parameters if available

        :param string stage_name: name of stage to start
        :param string mission_id: unique identifier of mission currently being completed
        :param int retry: number of retries in the case of failures, (429 responses are common with complex graphs)
        default is 3
        :param bool ignore_dependencies: if set to True, the stage will be allowed to start regardless of the state of
                                         all upstream stages dependencies
        :returns dict: Houston response {"next": list(string), "complete": bool, "params": dict()}
        """

        payload = {"state": "started", "ignoreDependencies": ignore_dependencies}
        status_code, json_response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions/" + mission_id + "/stages/" + stage_name,
            data=json.dumps(payload),
            retry=retry,
        )

        return json_response

    def end_stage(self, stage_name, mission_id, retry=3, ignore_dependencies=False):
        """Ends a Houston stage, returns downstream stages and available parameters

        :param string stage_name: name of stage which has finished
        :param string mission_id: unique identifier of mission currently being completed
        :param int retry: number of retry attempts
        :param bool ignore_dependencies: if set to True, all downstream stages dependant on this stage will be ignored,
                                         effectively ending the mission early
        :returns dict: Houston response {"next": list(string), "complete": bool, "params": dict(stage_name: dict())}
                       params contains multiple stage's parameters stored by stage name as keys
                       (e.g. {"next": "stage-two", "complete": False, "params": {"stage-two": dict()}})
        """

        payload = {"state": "finished", "ignoreDependencies": ignore_dependencies}
        status_code, json_response = self.interface_request.request(
            "post",
            uri=self.base_url + "/missions/" + mission_id + "/stages/" + stage_name,
            data=json.dumps(payload),
            retry=retry,
        )

        return json_response



"""Houston Utilities for Google Cloud Platform

PubSub utils:
Allows user to create subscriber/publish message according to the plan stage options, e.g.:

    res = h.end_stage("load-data", id)
    next_stage = res["next"]
    h.gcp.trigger(next_stage, id)

--> finds the topic it needs to use for the next_task from the stage parameters in the plan statement
--> sends a pub/sub message to the next task's topic

or:

    h.gcp.project = "my-project-1234"
    h.gcp.topic = "ps-topic-next-stage"

    res = h.end_stage("load-data", id)
    next_stage = res["next"]
    h.gcp.trigger_stage_via_pubsub(next_stage, id)

"""

import base64
import json
import os
from google.cloud import pubsub_v1
from houston.client import Houston


class GCPHouston(Houston):

    project = os.getenv("GCP_PROJECT", None)
    topic = None

    def call_stage_via_pubsub(self, response, mission_id):
        """Send stage details to Google Cloud Platform PubSub. Sends stage, mission_id, plan name as json in message
           body, parameters as attributes

           Message parameter must contain "psq" (PubSub Queue) key, this informs the function which topic is relevant
           to the task

           Blocks until PubSub message has been sent

        :param dict response: Response from Houston.end_stage
        """

        publisher_client = pubsub_v1.PublisherClient()
        if self.project is None:
            raise ValueError(
                "Project is not set. Use GCPHouston.project = '[PROJECT]' "
                "or set 'GCP_PROJECT' environment variable"
            )

        # for all available tasks - trigger qs
        for next_task in response["next"]:
            if next_task not in response["params"]:
                print(
                    "task: {next_task} does not have parameters, skipping".format(
                        next_task=next_task
                    )
                )
                continue
            if "psq" not in response["params"][next_task].keys():
                print(
                    "task: {next_task} does not have psq topic set, skipping".format(
                        next_task=next_task
                    )
                )
                continue

            task_parameters = response["params"][next_task]
            target_psq = task_parameters.pop("psq")

            data = json.dumps(
                {"stage": next_task, "mission_id": mission_id, "plan": self.plan}
            ).encode("utf-8")

            # make topic string
            topic = "projects/{project}/topics/{topic}".format(
                project=self.project, topic=target_psq
            )

            # json encode task param values
            # useful for decoding in PubSub subscriber
            for key, value in task_parameters.items():
                task_parameters[key] = json.dumps(value)

            if not task_parameters:
                future = publisher_client.publish(topic=topic, data=data)
                future.result()
            else:
                future = publisher_client.publish(
                    topic=topic, data=data, **task_parameters
                )
                future.result()

    @staticmethod
    def extract_stage_information(data):
        """Static method to extract stage information from sent PubSub message"""
        return json.loads(base64.b64decode(data))



from setuptools import setup, find_packages
from os import path

# Copy README.md to project description
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="houston-client",
    packages=find_packages(),
    description="Houston Python Client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="1.0.0",
    url="https://github.com/datasparq-intelligent-products/houston-python",
    author="James Watkinson, Matt Simmons & Ivan Nedjalkov",
    license="MIT",
    author_email="james.watkinson@datasparq.ai",
    install_requires=["requests==2.22.0", "requests[security]==2.22.0"],
    extras_require={"gcp": ["google-cloud-pubsub==1.2.0"]},
    keywords=["houston"],
)
